@@ -18,7 +18,7 @@
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * or visit http://www.gnu.org/copyleft/gpl.html
 *
-* 20010215 raf <raf@raf.org>
+* 20011109 raf <raf@raf.org>
 */
 
 /*
@@ -31,10 +31,13 @@ I<libslack(daemon)> - daemon module
 
     #include <slack/daemon.h>
 
+    typedef void daemon_config_parser_t(void *obj, const char *path, char *line, size_t lineno);
+
     int daemon_started_by_init(void);
     int daemon_started_by_inetd(void);
     int daemon_prevent_core(void);
     int daemon_revoke_privileges(void);
+    int daemon_become_user(uid_t uid, gid_t gid, char *user);
     char *daemon_absolute_path(const char *path);
     int daemon_path_is_safe(const char *path);
     void *daemon_parse_config(const char *path, void *obj, daemon_config_parser_t *parser);
@@ -57,9 +60,9 @@ tedious. These functions perform these tasks for you.
 #define _BSD_SOURCE /* For setgroups(2) and S_ISLNK(2) on Linux */
 #endif
 
+#include "config.h"
 #include "std.h"
 
-#include <signal.h>
 #include <fcntl.h>
 #include <pwd.h>
 #include <grp.h>
@@ -74,10 +77,12 @@ tedious. These functions perform these tasks for you.
 #include "lim.h"
 #include "fio.h"
 
-#ifdef NEEDS_SNPRINTF
+#ifndef HAVE_SNPRINTF
 #include "snprintf.h"
 #endif
 
+#ifndef TEST
+
 static struct
 {
 	pthread_mutex_t lock; /* Mutex lock for structure */
@@ -95,10 +100,10 @@ g =
 
 =item C<int daemon_started_by_init(void)>
 
-If this process was started by I<init(8)>, returns 1. If not, returns 0. If
-it was, we might be getting respawned so I<fork(2)> and I<exit(2)> would be
-a big mistake (and unnecessary anyway since there is no controlling
-terminal). The return value is cached so any subsequent calls are faster.
+If this process was started by I<init(8)>, returns 1. If not, returns C<0>.
+If it was, we might be getting respawned so I<fork(2)> and I<exit(2)> would
+be a big mistake (and unnecessary anyway since there is no controlling
+terminal). On error, returns C<-1> with C<errno> set appropriately.
 
 =cut
 
@@ -106,31 +111,18 @@ terminal). The return value is cached so any subsequent calls are faster.
 
 int daemon_started_by_init(void)
 {
-	static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
-	static int rc = -1;
-
-	if (rc == -1)
-	{
-		ptry(pthread_mutex_lock(&lock))
-
-		if (rc == -1)
-			rc = (getppid() == 1);
-
-		ptry(pthread_mutex_unlock(&lock))
-	}
-
-	return rc;
+	return (getppid() == 1);
 }
 
 /*
 
 =item C<int daemon_started_by_inetd(void)>
 
-If this process was started by I<inetd(8)>, returns 1. If not, returns 0. If
-it was, C<stdin>, C<stdout> and C<stderr> would be opened to a socket.
-Closing them would be a big mistake. We also would not need to I<fork(2)>
-and I<exit(2)> because there is no controlling terminal. The return value is
-cached so any subsequent calls are faster.
+If this process was started by I<inetd(8)>, returns C<1>. If not, returns
+C<0>. On error, returns C<-1> with C<errno> set appropriately. If it was,
+C<stdin>, C<stdout> and C<stderr> would be opened to a socket. Closing them
+would be a big mistake. We also would not need to I<fork(2)> and I<exit(2)>
+because there is no controlling terminal.
 
 =cut
 
@@ -138,32 +130,19 @@ cached so any subsequent calls are faster.
 
 int daemon_started_by_inetd(void)
 {
-	static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
-	static int rc = -1;
+	size_t optlen = sizeof(int);
+	int optval;
 
-	if (rc == -1)
-	{
-		size_t optlen = sizeof(int);
-		int optval;
-
-		ptry(pthread_mutex_lock(&lock))
-
-		if (rc == -1)
-			rc = (getsockopt(STDIN_FILENO, SOL_SOCKET, SO_TYPE, &optval, &optlen) == 0);
-
-		ptry(pthread_mutex_unlock(&lock))
-	}
-
-	return rc;
+	return (getsockopt(STDIN_FILENO, SOL_SOCKET, SO_TYPE, &optval, &optlen) == 0);
 }
 
 /*
 
 =item C<int daemon_prevent_core(void)>
 
-Prevents core files from being generated. This is used to prevent security
-holes in daemons run by root. On success, returns 0. On error, returns -1
-with C<errno> set appropriately.
+Prevents core files from being generated. This is used to prevent leaking
+sensitive information in daemons run by root. On success, returns C<0>. On
+error, returns C<-1> with C<errno> set appropriately.
 
 =cut
 
@@ -185,15 +164,16 @@ int daemon_prevent_core(void)
 
 =item C<int daemon_revoke_privileges(void)>
 
-Revokes suid and sgid privileges. Useful when your program does not require
-any special privileges and may become unsafe if incorrectly installed with
-special privileges. Also useful when your program only requires special
-privileges upon startup (e.g. binding to a privileged socket). Performs the
-following: Clears all supplementary groups. Sets the effective gid to the
-real gid if they differ. Sets the effective uid to the real uid if they
-differ. Checks that they no longer differ. Also closes /etc/passwd and
-/etc/group in case they were opened by root and give access to user and
-group passwords. On success, returns 0. On error, returns -1.
+Revokes setuid and setgid privileges. Useful when your program does not
+require any special privileges and may become unsafe if incorrectly
+installed with special privileges. Also useful when your program only
+requires special privileges upon startup (e.g. binding to a privileged
+socket). Performs the following: Sets the effective gid to the real gid if
+they differ. Checks that they no longer differ. Sets the effective uid to
+the real uid if they differ. Checks that they no longer differ. Also closes
+/etc/passwd and /etc/group in case they were opened by root and give access
+to user and group passwords. On success, returns C<0>. On error, returns
+C<-1> with C<errno> set appropriately.
 
 =cut
 
@@ -206,9 +186,6 @@ int daemon_revoke_privileges(void)
 	uid_t euid = geteuid();
 	gid_t egid = getegid();
 
-	if (euid == 0 && euid != uid && (setgroups(0, NULL) == -1 || getgroups(0, NULL) != 0))
-		return -1;
-
 	if (egid != gid && (setgid(gid) == -1 || getegid() != getgid()))
 		return -1;
 
@@ -223,14 +200,45 @@ int daemon_revoke_privileges(void)
 
 /*
 
+=item C<int daemon_become_user(uid_t uid, gid_t gid, char *user)>
+
+Changes the owner and group of the process to C<uid> and C<gid>
+respectively. If C<user> is not null, the supplementary group list will be
+initialised with I<initgroups()>. Otherwise, the supplementary group list
+will be cleared of all groups. On success, returns 0. On error, returns -1.
+Only root can use this function.
+
+=cut
+
+*/
+
+int daemon_become_user(uid_t uid, gid_t gid, char *user)
+{
+	if (setgroups(0, NULL) == -1 || getgroups(0, NULL) != 0)
+		return -1;
+
+	if (setgid(gid) == -1 || getgid() != gid || getegid() != gid)
+		return -1;
+
+	if (user && initgroups(user, gid) == -1)
+		return -1;
+
+	if (setuid(uid) == -1 || getuid() != uid || geteuid() != uid)
+		return -1;
+
+	return 0;
+}
+
+/*
+
 =item C<char *daemon_absolute_path(const char *path)>
 
 Returns C<path> converted into an absolute path. Cleans up any C<.> and
 C<..> and C<//> and trailing C</> found in the returned path. Note that the
 returned path looks canonical but isn't because symbolic links are not
 followed and expanded. It is the caller's responsibility to deallocate the
-path returned with C<mem_release(3)> or C<free(3)>. On success, returns the
-absolute path. On error, returns C<-1> with C<errno> set appropriately.
+path returned with I<mem_release(3)> or I<free(3)>. On success, returns the
+absolute path. On error, returns C<null> with C<errno> set appropriately.
 
 =cut
 
@@ -243,10 +251,7 @@ char *daemon_absolute_path(const char *path)
 	char *p;
 
 	if (!path)
-	{
-		set_errno(EINVAL);
-		return NULL;
-	}
+		return set_errnull(EINVAL);
 
 	/* Make path absolute and mostly canonical (don't follow symbolic links) */
 
@@ -267,29 +272,24 @@ char *daemon_absolute_path(const char *path)
 		}
 
 		cwd_len = strlen(cwd);
-
 		if (cwd_len + 1 + strlen(path) >= lim)
 		{
 			mem_release(cwd);
-			set_errno(ENAMETOOLONG);
-			return NULL;
+			return set_errnull(ENAMETOOLONG);
 		}
 
 		rc = snprintf(cwd + cwd_len, lim - cwd_len, "%c%s", PATH_SEP, path);
-
 		if (rc == -1 || rc >= lim - cwd_len)
 		{
 			mem_release(cwd);
-			set_errno(ENAMETOOLONG);
-			return NULL;
+			return set_errnull(ENAMETOOLONG);
 		}
 
 		abs_path = cwd;
 	}
 	else
 	{
-		abs_path = mem_strdup(path);
-		if (!abs_path)
+		if (!(abs_path = mem_strdup(path)))
 			return NULL;
 	}
 
@@ -385,11 +385,10 @@ static int daemon_check_path(char *path, int level)
 			char *tmp;
 
 			lim = limit_path();
-			sym_linked = mem_create(lim, char);
-			if (!sym_linked)
+			if (!(sym_linked = mem_create(lim, char)))
 				return -1;
 
-			memset(sym_linked, nul, lim);
+			memset(sym_linked, 0, lim);
 
 			if (readlink(path, sym_linked, lim) == -1)
 			{
@@ -399,8 +398,7 @@ static int daemon_check_path(char *path, int level)
 
 			if (*sym_linked != PATH_SEP)
 			{
-				tmp = mem_create(lim, char);
-				if (!tmp)
+				if (!(tmp = mem_create(lim, char)))
 				{
 					mem_release(sym_linked);
 					return -1;
@@ -425,8 +423,7 @@ static int daemon_check_path(char *path, int level)
 
 			tmp = daemon_absolute_path(sym_linked);
 			mem_release(sym_linked);
-			sym_linked = tmp;
-			if (!sym_linked)
+			if (!(sym_linked = tmp))
 				return -1;
 
 			rc = daemon_check_path(sym_linked, level + 1);
@@ -474,11 +471,11 @@ int daemon_path_is_safe(const char *path)
 =item C<void *daemon_parse_config(const char *path, void *obj, daemon_config_parser_t *parser)>
 
 Parses the text configuration file named C<path>. Blank lines are ignored.
-Comments (C<`#'> to end of line) are ignored. Lines that end with C<`\'> are
+Comments (C<'#'> to end of line) are ignored. Lines that end with C<'\'> are
 joined with the following line. There may be whitespace and even a comment
-after the C<`\'> character but nothing else. The C<parser> function is
+after the C<'\'> character but nothing else. The C<parser> function is
 called with the client supplied C<obj>, the file name, the line and the line
-number as arguments. On success, returns C<obj>. On errors, returns C<NULL>
+number as arguments. On success, returns C<obj>. On error, returns C<null>
 (i.e. if the configuration file could not be read). Note: Don't parse config
 files unless they are "safe" as determined by I<daemon_path_is_safe()>.
 
@@ -508,8 +505,7 @@ void *daemon_parse_config(const char *path, void *obj, daemon_config_parser_t *p
 
 		/* Strip trailing comments */
 
-		end = strchr(start, '#');
-		if (end)
+		if ((end = strchr(start, '#')))
 			*end = nul;
 		else
 			end = start + strlen(start);
@@ -526,8 +522,7 @@ void *daemon_parse_config(const char *path, void *obj, daemon_config_parser_t *p
 
 		/* Perform line continuation */
 
-		cont_line = (end[-1] == '\\');
-		if (cont_line)
+		if ((cont_line = (end[-1] == '\\')))
 			--end;
 
 		length = strlen(line);
@@ -555,12 +550,12 @@ C<int daemon_pidfile(const char *name)>
 
 Creates a pid file for a daemon and locks it. The file has one line
 containing the process id of the daemon. The well-known locations for the
-file is defined in C<ROOT_PID_DIR> for I<root> (by default, C<"/var/run"> on
+file is defined in C<ROOT_PID_DIR> for root (by default, C<"/var/run"> on
 Linux and C<"/etc"> on Solaris) and C<USER_PID_DIR> for all other users
 (C<"/tmp"> by default). The name of the file is the name of the daemon
 (given by the name argument) followed by C<".pid">. The presence of this
 file will prevent two daemons with the same name from running at the same
-time. On success, returns 0. On error, returns -1 with C<errno> set
+time. On success, returns C<0>. On error, returns C<-1> with C<errno> set
 appropriately.
 
 */
@@ -573,7 +568,6 @@ static int daemon_pidfile(const char *name)
 	long path_len;
 	const char *pid_dir;
 	char *suffix = ".pid";
-	int rc;
 
 	path_len = limit_path();
 	pid_dir = (getuid()) ? USER_PID_DIR : ROOT_PID_DIR;
@@ -582,14 +576,9 @@ static int daemon_pidfile(const char *name)
 		return set_errno(ENAMETOOLONG);
 
 	if (!g.pidfile && !(g.pidfile = mem_create(path_len, char)))
-		return set_errno(ENOMEM);
+		return -1;
 
-	rc = snprintf(g.pidfile, path_len, "%s%c%s%s", pid_dir, PATH_SEP, name, suffix);
-	if (rc == -1 || rc >= path_len)
-	{
-		mem_destroy(&g.pidfile);
-		return set_errno(ENOSPC);
-	}
+	snprintf(g.pidfile, path_len, "%s%c%s%s", pid_dir, PATH_SEP, name, suffix);
 
 	/* This is broken over NFS (Linux). So pidfiles must reside locally. */
 
@@ -625,12 +614,7 @@ static int daemon_pidfile(const char *name)
 
 	/* It wasn't locked. Now we have it locked, store our pid. */
 
-	rc = snprintf(pid, 32, "%d\n", getpid());
-	if (rc == -1 || rc >= 32)
-	{
-		mem_destroy(&g.pidfile);
-		return set_errno(ENOSPC);
-	}
+	snprintf(pid, 32, "%d\n", getpid());
 
 	if (write(pid_fd, pid, strlen(pid)) != strlen(pid))
 	{
@@ -674,13 +658,20 @@ Start a new process session.
 Under I<SVR4>, background the process again to lose process session
 leadership. This prevents the process from ever gaining a controlling
 terminal. This only happens when C<SVR4> is defined and
-C<NO_EXTRA_SVR4_FORK> is not defined when I<libslack> is compiled.
+C<NO_EXTRA_SVR4_FORK> is not defined when I<libslack> is compiled. Before
+doing this, ignore C<SIGHUP> because when the session leader terminates, all
+processes in the foreground process group are sent a C<SIGHUP> signal
+(apparently). Note that this code may not execute (e.g. when started by
+I<init(8)> or I<inetd(8)> or when either C<SVR4> was not defined or
+C<NO_EXTRA_SVR4_FORK> was defined when I<libslack> was compiled). This means
+that the client can't make any assumptions about the C<SIGHUP> handler when
+I<daemon_init()> returns.
 
 =back
 
 =item *
 
-Change to the root directory so as not to hamper umounts.
+Change directory to the root directory so as not to hamper umounts.
 
 =item *
 
@@ -700,13 +691,16 @@ invoked by I<inetd(8)>.
 
 =item *
 
-If C<name> is non-NULL, create and lock a file containing the process id of
+If C<name> is non-null, create and lock a file containing the process id of
 the process. The presence of this locked file prevents two instances of a
-daemon with the same name from running at the same time.
+daemon with the same name from running at the same time. The default
+location of the pidfile is C</var/run> on Linux and C</etc> on Solaris for
+I<root> or C</tmp> for ordinary users.
 
 =back
 
-On success, returns 0. On error, returns -1 with C<errno> set appropriately.
+On success, returns C<0>. On error, returns C<-1> with C<errno> set
+appropriately.
 
 =cut
 
@@ -727,21 +721,41 @@ int daemon_init(const char *name)
 	{
 		/*
 		** Background the process.
-		** Lose process group leadership.
+		** Lose process session/group leadership.
 		*/
 
 		if ((pid = fork()) == -1)
 			return -1;
 
 		if (pid)
-			exit(0);
+			exit(EXIT_SUCCESS);
 
 		/* Become a process session leader. */
+		/* This can only fail when we're already a session leader. */
 
 		setsid();
 
 #ifndef NO_EXTRA_SVR4_FORK
 #ifdef SVR4
+		/*
+		** Ignore SIGHUP because when the sesson leader terminates (which is
+		** about to happen), all processes in the foreground process group
+		** are sent the SIGHUP signal (apparently). It is expected that
+		** client's will set their own SIGHUP handler after the call to
+		** daemon_init() if necessary.
+		*/
+
+		{
+			struct sigaction act[1];
+
+			act->sa_handler = SIG_IGN;
+			sigemptyset(&act->sa_mask);
+			act->sa_flags = 0;
+
+			if (sigaction(SIGHUP, act, NULL) == -1)
+				return -1;
+		}
+
 		/*
 		** Lose process session leadership
 		** to prevent gaining a controlling
@@ -752,7 +766,7 @@ int daemon_init(const char *name)
 			return -1;
 
 		if (pid)
-			exit(0);
+			exit(EXIT_SUCCESS);
 #endif
 #endif
 	}
@@ -770,6 +784,10 @@ int daemon_init(const char *name)
 	** We need to close all open file descriptors. Check how
 	** many file descriptors we have (If indefinite, a usable
 	** number (1024) will be returned).
+	**
+	** Flaw: If many files were opened and then this limit
+	** was reduced to below the highest file descriptor,
+	** we may not close all file descriptors.
 	*/
 
 	if ((nopen = limit_open()) == -1)
@@ -884,46 +902,31 @@ calls. See their manual pages.
 
 =over 4
 
+=item EINVAL
+
+An argument was invalid (e.g. C<null>).
+
 =item ENAMETOOLONG
 
-The C<name> passed to I<daemon_init()> resulted in a path name that is
-too long for the intended filesystem.
+The C<name> passed to I<daemon_init()> or I<daemon_path_is_safe()> resulted
+in a path name that is too long for the intended filesystem.
 
-=item ENOMEM
+=item ELOOP
 
-I<daemon_init()> failed to allocate memory for the the pid file's path.
+I<daemon_path_is_safe()> recursed too deeply (16 levels).
 
 =back
 
 =head1 EXAMPLE
 
     #include <stdio.h>
-
     #include <unistd.h>
     #include <syslog.h>
-    #include <signal.h>
 
-    #include <slack/daemon.h>
-    #include <slack/prog.h>
-    #include <slack/sig.h>
+    #include <slack/lib.h>
 
-    void hup(int signo)
-    {
-        // reload config file...
-    }
-
-    void term(int signo)
-    {
-        daemon_close();
-        exit(0);
-    }
-
-    void do_stuff()
-    {
-        // do stuff...
-        kill(getpid(), SIGTERM);
-        signal_handle_all();
-    }
+    const char * const config_fname = "/etc/fstab";
+    List *config = NULL;
 
     void fstab_parser(void *obj, const char *path, char *line, size_t lineno)
     {
@@ -933,24 +936,52 @@ I<daemon_init()> failed to allocate memory for the the pid file's path.
         if (sscanf(line, "%s %s %s %s %d %d", device, mount, fstype, opts, &freq, &passno) != 6)
             fprintf(stderr, "Syntax error in %s (line %d): %s\n", path, lineno, line);
         else
+        {
+            char *copy;
+
             printf("%s %s %s %s %d %d\n", device, mount, fstype, opts, freq, passno);
+
+            if (!(copy = mem_strdup(line)))
+                fprintf(stderr, "out of memory\n");
+            else if (!list_append(config, copy))
+                fprintf(stderr, "failed to add line %d to config\n", lineno);
+        }
+    }
+
+    void hup(int signo)
+    {
+        list_remove_range(config, 0, -1);
+        daemon_parse_config(config_fname, config, fstab_parser);
+    }
+
+    void term(int signo)
+    {
+        daemon_close();
+        exit(EXIT_SUCCESS);
+    }
+
+    void do_stuff()
+    {
+        // do stuff...
+        kill(getpid(), SIGTERM);
+        signal_handle_all();
     }
 
     int main(int ac, char **av)
     {
-        const char * const config = "/etc/fstab";
-
         if (daemon_revoke_privileges() == -1 ||
             daemon_prevent_core() == -1 ||
-            daemon_parse_config(config, NULL, fstab_parser) == NULL ||
+            daemon_path_is_safe(config_fname) != 1 ||
+            (config = list_create(free)) == NULL ||
+            daemon_parse_config(config_fname, config, fstab_parser) == NULL ||
             daemon_init(prog_basename(*av)) == -1 ||
             signal_set_handler(SIGHUP, 0, hup) == -1 ||
-            signal_set_handler(SIGTERM, 0, term) == -1 ||
-            daemon_path_is_safe(config) != 1)
+            signal_set_handler(SIGTERM, 0, term) == -1)
             return 1;
 
         do_stuff();
-        return 0; // unreached
+
+        return EXIT_SUCCESS; // unreached
     }
 
 =head1 MT-Level
@@ -960,7 +991,7 @@ MT-Safe
 =head1 BUGS
 
 It is possible to obtain a controlling terminal under I<BSD> (and even under
-I<SVR4> if C<SVR4> was not defined or C<NO_EXTRA_SVR4_FORK> was defined when
+I<SVR4> if I<SVR4> was not defined or C<NO_EXTRA_SVR4_FORK> was defined when
 I<libslack> is compiled). If anything calls I<open(2)> on a terminal device
 without the C<O_NOCTTY> flag, the process doing so will obtain a controlling
 terminal.
@@ -975,13 +1006,13 @@ C<ROOT_PID_DIR> and C<USER_PID_DIR> to both be C</tmp>.
 The exclusive creation and locking of the pidfile doesn't work correctly
 over NFS on Linux so pidfiles must reside locally.
 
-I<daemon_path_is_safe()> ignores ACLs (so does I<sendmail>). It should
+I<daemon_path_is_safe()> ignores ACLs (so does I<sendmail(8)>). It should
 probably treat a path as unsafe if there are any ACLs (allowing extra
 access) along the path.
 
-The functions C<daemon_prevent_core()>, C<daemon_revoke_privileges()>,
-C<daemon_absolute_path()>, C<daemon_path_is_safe()> and
-C<daemon_parse_config()> should probably all have the C<daemon_> prefix
+The functions I<daemon_prevent_core()>, I<daemon_revoke_privileges()>,
+I<daemon_become_user()>, I<daemon_absolute_path()>, I<daemon_path_is_safe()>
+and I<daemon_parse_config()> should probably all have the I<daemon_> prefix
 removed from their names. Their use is more general than just in daemons.
 
 =head1 SEE ALSO
@@ -989,20 +1020,34 @@ removed from their names. Their use is more general than just in daemons.
 L<libslack(3)|libslack(3)>,
 L<daemon(1)|daemon(1)>,
 L<init(8)|init(8)>,
-L<inetd(8)|inetd(8)>
+L<inetd(8)|inetd(8)>,
+L<fork(2)|fork(2)>,
+L<umask(2)|umask(2)>,
+L<setsid(2)|setsid(2)>,
+L<chdir(2)|chdir(2)>,
+L<setrlimit(2)|setrlimit(2)>,
+L<setgid(2)|setgid(2)>,
+L<setuid(2)|setuid(2)>,
+L<setgroups(2)|setgroups(2)>,
+L<initgroups(3)|initgroups(3)>,
+L<endpwent(3)|endpwent(3)>,
+L<endgrent(3)|endgrent(3)>
 
 =head1 AUTHOR
 
-20010215 raf <raf@raf.org>
+20011109 raf <raf@raf.org>
 
 =cut
 
 */
 
+#endif
+
 #ifdef TEST
 
 #include <syslog.h>
-#include <wait.h>
+
+#include <sys/wait.h>
 
 #include "msg.h"
 #include "prog.h"
@@ -1131,7 +1176,7 @@ static void parse_test2(void *obj, const char *path, char *line, size_t lineno)
 	Data2 *data2 = (Data2 *)obj;
 	char word[8][BUFSIZ];
 	int words;
-	
+
 	words = sscanf(line, "%s %s %s %s %s %s %s %s", word[0], word[1], word[2], word[3], word[4], word[5], word[6], word[7]);
 
 	for (data2->j = 0; data2->j < words; ++data2->j)
@@ -1156,13 +1201,15 @@ static void parse_test2(void *obj, const char *path, char *line, size_t lineno)
 void term(int signo)
 {
 	daemon_close();
-	exit(0);
+	exit(EXIT_SUCCESS);
 }
 
 int main(int ac, char **av)
 {
 	const char *config_name;
 	char *cwd;
+	char *core = "core";
+	char *core2 = "daemon.core"; /* OpenBSD */
 	int facility = LOG_DAEMON | LOG_ERR;
 	pid_t pid;
 	int rc;
@@ -1170,6 +1217,13 @@ int main(int ac, char **av)
 	gid_t gid, egid;
 	int no_privileges = 0;
 	int not_safe = 0;
+	int not_root = 0;
+
+	if (ac == 2 && !strcmp(av[1], "help"))
+	{
+		printf("usage: %s\n", *av);
+		return EXIT_SUCCESS;
+	}
 
 	printf("Testing: daemon\n");
 
@@ -1183,7 +1237,8 @@ int main(int ac, char **av)
 
 	/* Test daemon_prevent_core() */
 
-	unlink("core");
+	unlink(core);
+	unlink(core2);
 
 	switch (pid = fork())
 	{
@@ -1215,11 +1270,16 @@ int main(int ac, char **av)
 				break;
 			}
 
-			if (WIFEXITED(status) && WEXITSTATUS(status) != 0)
+#ifndef WCOREDUMP
+#define WCOREDUMP 0
+#endif
+
+			if (WIFEXITED(status) && WEXITSTATUS(status))
 				++errors;
-			else if (WCOREDUMP(status) || stat("core", statbuf) == 0)
+			else if (WCOREDUMP(status) || stat(core, statbuf) == 0 || stat(core2, statbuf) == 0)
 				++errors, printf("Test3: child dumped core\n");
-			unlink("core");
+			unlink(core);
+			unlink(core2);
 		}
 	}
 
@@ -1235,6 +1295,60 @@ int main(int ac, char **av)
 	else if (daemon_revoke_privileges() == -1 || geteuid() != getuid() || getegid() != getgid())
 		++errors, printf("Test4: daemon_revoke_privileges() failed: %s\n", strerror(errno));
 
+	/* Test daemon_become_user() if possible */
+
+	if (uid)
+		not_root = 1;
+	else
+	{
+		switch (pid = fork())
+		{
+			case -1:
+			{
+				++errors, printf("Test5: Failed to run test: fork: %s\n", strerror(errno));
+				break;
+			}
+
+			case 0:
+			{
+				errno = 0;
+				if (daemon_become_user(1, 1, NULL) == -1)
+				{
+					printf("Test5: daemon_become_user(1, 1, NULL) failed: %s\n", strerror(errno));
+					return 1;
+				}
+
+				if (geteuid() != 1 || getuid() != 1 || getegid() != 1 || getgid() != 1)
+				{
+					printf("Test5: daemon_become_user(1, 1, NULL) failed: euid/egid = %d/%d, uid/gid = %d/%d\n", (int)geteuid(), (int)getegid(), (int)getuid(), (int)getgid());
+					return 1;
+				}
+
+				if (getgroups(0, NULL) != 0)
+				{
+					printf("Test5: daemon_become_user(1, 1, NULL) failed: getgroups() = %d (not 0)\n", getgroups(0, NULL));
+					return 1;
+				}
+
+				return 0;
+			}
+
+			default:
+			{
+				int status;
+
+				if (waitpid(pid, &status, 0) == -1)
+				{
+					printf("Test5: Failed to evaluate test: waitpid: %s\n", strerror(errno));
+					break;
+				}
+
+				if (WIFEXITED(status) && WEXITSTATUS(status))
+					++errors;
+			}
+		}
+	}
+
 	/* Test daemon_absolute_path() */
 
 #define TEST_ABSOLUTE_PATH(i, path, abs_path) \
@@ -1244,7 +1358,12 @@ int main(int ac, char **av)
 			++errors, printf("Test%d: absolute_path(%s) failed (%s)\n", (i), (path), strerror(errno)); \
 		else if (strcmp(result, (abs_path))) \
 		{ \
+			struct stat result_status[1], abs_status[1]; \
 			++errors, printf("Test%d: absolute_path(%s) failed (was %s, not %s)\n", (i), (path), result, (abs_path)); \
+			printf("\n"); \
+			if (stat(result, result_status) != -1 && stat(abs_path, abs_status) != -1 && result_status->st_ino == abs_status->st_ino) \
+				printf("        But they have the same inode (%d)\n", (int)abs_status->st_ino); \
+				printf("        Does your pwd return canonical paths?\n\n"); \
 			free(result); \
 		} \
 	}
@@ -1252,32 +1371,32 @@ int main(int ac, char **av)
 	/* We must be in a safe, writable directory to test relative paths */
 
 	if (!(cwd = mem_create(limit_path(), char)))
-		++errors, printf("Test5: Failed to run test: mem_create: %s\n", strerror(errno));
+		++errors, printf("Test6: Failed to run test: mem_create: %s\n", strerror(errno));
 	else if (!getcwd(cwd, limit_path()))
-		++errors, printf("Test5: Failed to run test: getcwd: %s\n", strerror(errno));
+		++errors, printf("Test6: Failed to run test: getcwd: %s\n", strerror(errno));
 	else if (chdir("/etc") == -1)
-		++errors, printf("Test5: Failed to run test: chdir: %s\n", strerror(errno));
+		++errors, printf("Test6: Failed to run test: chdir: %s\n", strerror(errno));
 	else
 	{
-		TEST_ABSOLUTE_PATH(5, ".", "/etc")
-		TEST_ABSOLUTE_PATH(6, "..", "/")
-		TEST_ABSOLUTE_PATH(7, "/", "/")
-		TEST_ABSOLUTE_PATH(8, "/etc/passwd", "/etc/passwd")
-		TEST_ABSOLUTE_PATH(9, "/.", "/")
-		TEST_ABSOLUTE_PATH(10, "/..", "/")
-		TEST_ABSOLUTE_PATH(11, "/./etc", "/etc")
-		TEST_ABSOLUTE_PATH(12, "/../etc", "/etc")
-		TEST_ABSOLUTE_PATH(13, "/etc/.././.././../usr", "/usr")
-		TEST_ABSOLUTE_PATH(14, "../../../../../etc/././../etc/./.././etc", "/etc")
-		TEST_ABSOLUTE_PATH(15, "././../../../../../etc/././.", "/etc")
-		TEST_ABSOLUTE_PATH(16, "/etc/./sysconfig/./network-scripts/../blog/..", "/etc/sysconfig")
-		TEST_ABSOLUTE_PATH(17, "/etc/./sysconfig/./network-scripts/../blog/../..", "/etc")
-		TEST_ABSOLUTE_PATH(18, "passwd", "/etc/passwd")
-		TEST_ABSOLUTE_PATH(19, "passwd/", "/etc/passwd")
-		TEST_ABSOLUTE_PATH(20, "passwd////", "/etc/passwd")
-		TEST_ABSOLUTE_PATH(21, "///////////////", "/")
-		TEST_ABSOLUTE_PATH(22, "///////etc////////", "/etc")
-		TEST_ABSOLUTE_PATH(23, "//////./.././..////..//", "/")
+		TEST_ABSOLUTE_PATH(6, ".", "/etc")
+		TEST_ABSOLUTE_PATH(7, "..", "/")
+		TEST_ABSOLUTE_PATH(8, "/", "/")
+		TEST_ABSOLUTE_PATH(9, "/etc/passwd", "/etc/passwd")
+		TEST_ABSOLUTE_PATH(10, "/.", "/")
+		TEST_ABSOLUTE_PATH(11, "/..", "/")
+		TEST_ABSOLUTE_PATH(12, "/./etc", "/etc")
+		TEST_ABSOLUTE_PATH(13, "/../etc", "/etc")
+		TEST_ABSOLUTE_PATH(14, "/etc/.././.././../usr", "/usr")
+		TEST_ABSOLUTE_PATH(15, "../../../../../etc/././../etc/./.././etc", "/etc")
+		TEST_ABSOLUTE_PATH(16, "././../../../../../etc/././.", "/etc")
+		TEST_ABSOLUTE_PATH(17, "/etc/./sysconfig/./network-scripts/../blog/..", "/etc/sysconfig")
+		TEST_ABSOLUTE_PATH(18, "/etc/./sysconfig/./network-scripts/../blog/../..", "/etc")
+		TEST_ABSOLUTE_PATH(19, "passwd", "/etc/passwd")
+		TEST_ABSOLUTE_PATH(20, "passwd/", "/etc/passwd")
+		TEST_ABSOLUTE_PATH(21, "passwd////", "/etc/passwd")
+		TEST_ABSOLUTE_PATH(22, "///////////////", "/")
+		TEST_ABSOLUTE_PATH(23, "///////etc////////", "/etc")
+		TEST_ABSOLUTE_PATH(24, "//////./.././..////..//", "/")
 		chdir(cwd);
 	}
 
@@ -1288,14 +1407,19 @@ int main(int ac, char **av)
 		int rc; \
 		errno = 0; \
 		if ((rc = daemon_path_is_safe(path)) != (safe)) \
+		{ \
+			struct stat status[1]; \
 			++errors, printf("Test%d: daemon_path_is_safe(%s) failed (ret %d, not %d) %s\n", (i), (path), rc, (safe), errno ? strerror(errno) : ""); \
+			if (stat(ROOT_DIR, status) != -1 && status->st_mode & (S_IWGRP | S_IWOTH)) \
+				printf("\n        No Wonder! Your %s directory is %s writable!!!\n\n", ROOT_DIR, (status->st_mode & S_IWOTH) ? "world" : "group"); \
+		} \
 		else if (rc == -1 && errno != (err)) \
 			++errors, printf("Test%d: daemon_path_is_safe(%s) failed (errno was %d, not %d)\n", (i), (path), errno, (err)); \
 	}
 
-	TEST_PATH_IS_SAFE(24, "/etc/passwd", 1, 0)
-	TEST_PATH_IS_SAFE(25, "/tmp", 0, 0)
-	TEST_PATH_IS_SAFE(26, "/nonexistent-path", -1, ENOENT)
+	TEST_PATH_IS_SAFE(25, "/etc/passwd", 1, 0)
+	TEST_PATH_IS_SAFE(26, "/tmp", 0, 0)
+	TEST_PATH_IS_SAFE(27, "/nonexistent-path", -1, ENOENT)
 
 	if (daemon_path_is_safe(".") != 1)
 		not_safe = 1;
@@ -1309,74 +1433,74 @@ int main(int ac, char **av)
 		/* Test absolute link from safe directory to unsafe directory */
 
 		if ((fd = open(sym_linked, O_CREAT | O_EXCL, S_IRUSR | S_IWUSR)) == -1)
-			++errors, printf("Test27: Failed to run test: open(%s) failed %s\n", sym_linked, strerror(errno));
+			++errors, printf("Test28: Failed to run test: open(%s) failed %s\n", sym_linked, strerror(errno));
 		else
 		{
 			close(fd);
-			TEST_PATH_IS_SAFE(27, sym_linked, 0, 0)
+			TEST_PATH_IS_SAFE(28, sym_linked, 0, 0)
 
 			if (symlink(sym_linked, sym_link) == -1)
-				++errors, printf("Test27: Failed to run test: symlink(%s, %s) failed %s\n", sym_linked, sym_link, strerror(errno));
+				++errors, printf("Test28: Failed to run test: symlink(%s, %s) failed %s\n", sym_linked, sym_link, strerror(errno));
 			else
 			{
-				TEST_PATH_IS_SAFE(27, sym_link, 0, 0)
+				TEST_PATH_IS_SAFE(28, sym_link, 0, 0)
 
 				if (unlink(sym_link) == -1)
-					++errors, printf("Test27: Failed to unlink(%s): %s\n", sym_link, strerror(errno));
+					++errors, printf("Test28: Failed to unlink(%s): %s\n", sym_link, strerror(errno));
 			}
 
 			if (unlink(sym_linked) == -1)
-				++errors, printf("Test27: Failed to unlink(%s): %s\n", sym_linked, strerror(errno));
+				++errors, printf("Test28: Failed to unlink(%s): %s\n", sym_linked, strerror(errno));
 		}
 
 		/* Test relative symbolic link from safe directory to safe directory */
 
 		if (mkdir("safedir", S_IRUSR | S_IWUSR | S_IXUSR) == -1)
-			++errors, printf("Test28: Failed to run test: mkdir(%s) failed: %s\n", "safedir", strerror(errno));
+			++errors, printf("Test29: Failed to run test: mkdir(%s) failed: %s\n", "safedir", strerror(errno));
 		else
 		{
 			if (symlink("..", "safedir/safelink") == -1)
-				++errors, printf("Test28: symlink(.., safedir/safelink) failed: %s\n", strerror(errno));
+				++errors, printf("Test29: symlink(.., safedir/safelink) failed: %s\n", strerror(errno));
 			else
 			{
-				TEST_PATH_IS_SAFE(28, "safedir/safelink", 1, 0)
+				TEST_PATH_IS_SAFE(29, "safedir/safelink", 1, 0)
 
 				if (unlink("safedir/safelink") == -1)
-					++errors, printf("Test28: Failed to unlink(safedir/safelink): %s\n", strerror(errno));
+					++errors, printf("Test29: Failed to unlink(safedir/safelink): %s\n", strerror(errno));
 			}
 
 			if (rmdir("safedir") == -1)
-				++errors, printf("Test28: Failed to rmdir(safedir): %s\n", strerror(errno));
+				++errors, printf("Test29: Failed to rmdir(safedir): %s\n", strerror(errno));
 		}
 
 		/* Test relative symbolic link from safe directory to unsafe directory */
 
 		if (mkdir("safedir", S_IRUSR | S_IWUSR | S_IXUSR) == -1)
-			++errors, printf("Test29: Failed to run test: mkdir(safedir) failed: %s\n", strerror(errno));
+			++errors, printf("Test30: Failed to run test: mkdir(safedir) failed: %s\n", strerror(errno));
 		else
 		{
 			mask = umask((mode_t)0);
 
 			if (mkdir("unsafedir", S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP) == -1)
-				++errors, printf("Test29: Failed to run test: mkdir(unsafedir) failed: %s\n", strerror(errno));
+				++errors, printf("Test30: Failed to run test: mkdir(unsafedir) failed: %s\n", strerror(errno));
 			else
 			{
 				if (symlink("../unsafedir", "safedir/unsafelink") == -1)
-					++errors, printf("Test29: symlink(../unsafedir, safedir/unsafelink) failed: %s\n", strerror(errno));
+					++errors, printf("Test30: symlink(../unsafedir, safedir/unsafelink) failed: %s\n", strerror(errno));
 				else
 				{
-					TEST_PATH_IS_SAFE(28, "safedir/unsafelink", 0, 0)
+					TEST_PATH_IS_SAFE(30, "safedir/unsafelink", 0, 0)
 
 					if (unlink("safedir/unsafelink") == -1)
-						++errors, printf("Test29: Failed to unlink(safedir/unsafelink): %s\n", strerror(errno));
+						++errors, printf("Test30: Failed to unlink(safedir/unsafelink): %s\n", strerror(errno));
 				}
 
 				if (rmdir("unsafedir") == -1)
-					++errors, printf("Test29: Failed to rmdir(unsafedir): %s\n", strerror(errno));
+					++errors, printf("Test30: Failed to rmdir(unsafedir): %s\n", strerror(errno));
 			}
 
 			if (rmdir("safedir") == -1)
-				++errors, printf("Test29: Failed to rmdir(safedir): %s\n", strerror(errno));
+				++errors, printf("Test30: Failed to rmdir(safedir): %s\n", strerror(errno));
 
 			umask(mask);
 		}
@@ -1386,29 +1510,29 @@ int main(int ac, char **av)
 		mask = umask((mode_t)0);
 
 		if (mkdir("unsafedir", S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP) == -1)
-			++errors, printf("Test30: Failed to run test: mkdir(unsafedir) failed: %s\n", strerror(errno));
+			++errors, printf("Test31: Failed to run test: mkdir(unsafedir) failed: %s\n", strerror(errno));
 		else
 		{
 			if (symlink("unsafedir", "unsafelink") == -1)
-				++errors, printf("Test30: symlink(../unsafedir, unsafelink) failed: %s\n", strerror(errno));
+				++errors, printf("Test31: symlink(../unsafedir, unsafelink) failed: %s\n", strerror(errno));
 			else
 			{
 				if (mkdir("unsafelink/unsafedir", S_IRUSR | S_IWUSR | S_IXUSR) == -1)
-					++errors, printf("Test30: Failed to run test: mkdir(unsafelink/unsafedir) failed: %s\n", strerror(errno));
+					++errors, printf("Test31: Failed to run test: mkdir(unsafelink/unsafedir) failed: %s\n", strerror(errno));
 				else
 				{
-					TEST_PATH_IS_SAFE(30, "unsafelink/unsafedir", 0, 0)
+					TEST_PATH_IS_SAFE(31, "unsafelink/unsafedir", 0, 0)
 
 					if (rmdir("unsafelink/unsafedir") == -1)
-						++errors, printf("Test30: Failed to rmdir(unsafelink/unsafedir): %s\n", strerror(errno));
+						++errors, printf("Test31: Failed to rmdir(unsafelink/unsafedir): %s\n", strerror(errno));
 				}
 
 				if (unlink("unsafelink") == -1)
-					++errors, printf("Test30: Failed to unlink(unsafelink): %s\n", strerror(errno));
+					++errors, printf("Test31: Failed to unlink(unsafelink): %s\n", strerror(errno));
 			}
 
 			if (rmdir("unsafedir") == -1)
-				++errors, printf("Test30: Failed to rmdir(unsafedir): %s\n", strerror(errno));
+				++errors, printf("Test31: Failed to rmdir(unsafedir): %s\n", strerror(errno));
 		}
 
 		umask(mask);
@@ -1418,23 +1542,23 @@ int main(int ac, char **av)
 
 	config_name = "daemon_parse_config.testfile";
 
-	if (config_test1(31, config_name))
+	if (config_test1(32, config_name))
 	{
 		int errors_save = errors;
-		data1->test = 31;
+		data1->test = 32;
 		daemon_parse_config(config_name, data1, parse_test1);
 		if (errors == errors_save && data1->i != final_pair)
-			++errors, printf("Test31: failed to parse entire config file\n");
+			++errors, printf("Test32: failed to parse entire config file\n");
 		unlink(config_name);
 	}
 
-	if (config_test2(32, config_name))
+	if (config_test2(33, config_name))
 	{
 		int errors_save = errors;
-		data2->test = 32;
+		data2->test = 33;
 		daemon_parse_config(config_name, data2, parse_test2);
 		if (errors == errors_save && (data2->i != final_line || data2->j != final_word))
-			++errors, printf("Test32: failed to parse entire config file\n");
+			++errors, printf("Test33: failed to parse entire config file\n");
 		unlink(config_name);
 	}
 
@@ -1444,7 +1568,7 @@ int main(int ac, char **av)
 	{
 		case -1:
 		{
-			++errors, printf("Test33: Failed to run test: fork: %s\n", strerror(errno));
+			++errors, printf("Test34: Failed to run test: fork: %s\n", strerror(errno));
 			break;
 		}
 
@@ -1452,33 +1576,30 @@ int main(int ac, char **av)
 		{
 			if (signal_set_handler(SIGTERM, 0, term) == -1)
 			{
-				syslog(facility, "%s: Test33: signal_set_handler() failed: %s", *av, strerror(errno));
-				return 1;
+				syslog(facility, "%s: Test34: signal_set_handler(SIGTERM) failed: %s", *av, strerror(errno));
+				_exit(EXIT_FAILURE);
 			}
 
 			if (daemon_init(prog_basename(*av)) == -1)
 			{
-				syslog(facility, "%s: Test33: daemon_init() failed: %s", *av, strerror(errno));
-			}
-			else
-			{
-				syslog(facility, "%s succeeded", *av);
-
-				if (kill(getpid(), SIGTERM) == -1)
-					syslog(facility, "%s: Test33: kill(%d, SIGTERM) failed: %s", *av, pid, strerror(errno));
-
-				signal_handle_all();
-
-				/*
-				** We can only get here if the signal hasn't arrived yet.
-				** If so, exit anyway.
-				*/
-
-				syslog(facility, "%s: Test33: signal_handle_all() failed", *av);
-				_exit(0);
+				syslog(facility, "%s: Test34: daemon_init(\"%s\") failed: %s", *av, prog_basename(*av), strerror(errno));
+				_exit(EXIT_FAILURE);
 			}
 
-			return 0; /* unreached */
+			syslog(facility, "%s succeeded", *av);
+
+			if (kill(getpid(), SIGTERM) == -1)
+				syslog(facility, "%s: Test34: kill(%d, SIGTERM) failed: %s", *av, pid, strerror(errno));
+
+			signal_handle_all();
+
+			/*
+			** We can only get here if the signal hasn't arrived yet.
+			** If so, exit anyway.
+			*/
+
+			syslog(facility, "%s: Test34: signal_handle_all() failed", *av);
+			_exit(EXIT_SUCCESS);
 		}
 
 		default:
@@ -1487,17 +1608,17 @@ int main(int ac, char **av)
 
 			if (waitpid(pid, &status, 0) == -1)
 			{
-				printf("Test33: Failed to evaluate test: waitpid: %s\n", strerror(errno));
+				printf("Test34: Failed to evaluate test: waitpid: %s\n", strerror(errno));
 				break;
 			}
 
-			if (WIFEXITED(status) && WEXITSTATUS(status) != 0)
-				++errors, printf("Test33: Failed to run test: signal_set_handler() failed\n");
+			if (WIFEXITED(status) && WEXITSTATUS(status))
+				++errors, printf("Test34: Failed to run test: signal_set_handler() failed\n");
 		}
 	}
 
 	if (errors)
-		printf("%d/33 tests failed\n", errors);
+		printf("%d/34 tests failed\n", errors);
 	else
 		printf("All tests passed\n");
 
@@ -1519,7 +1640,14 @@ int main(int ac, char **av)
 		printf("    Rerun test from a safe directory (writable by the tester).\n");
 	}
 
-	return 0;
+	if (not_root)
+	{
+		printf("\n");
+		printf("    Note: Can't test daemon_become_user().\n");
+		printf("    Audit the code and rerun the test as root.\n");
+	}
+
+	return (errors == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
 }
 
 #endif
